/*
 * This file is part of John the Ripper password cracker,
 * Copyright (c) 2010 by Solar Designer
 *
 * MD4 OpenCL code is based on Alain Espinosa's OpenCL patches.
 */

#include <string.h>

#include "arch.h"
#include "params.h"
#include "path.h"
#include "common.h"
#include "formats.h"
#include "md4.h"

#define FORMAT_LABEL            "raw-md4"
#define FORMAT_NAME         "Raw MD4"
#define ALGORITHM_NAME          "32/" ARCH_BITS_STR
#define BENCHMARK_COMMENT       ""
#define BENCHMARK_LENGTH        -1
#define CIPHERTEXT_LENGTH       32
#define BINARY_SIZE         16
#define SALT_SIZE           0

#ifdef CL_VERSION_1_0

/*  !!! keep these two values synchronized !!! */
#define PLAINTEXT_LENGTH        15 // TODO: detect dynamically
#define MD4_NUM_KEYS         1024*2048

cl_platform_id platform;
cl_device_id devices;
cl_context context;
cl_program program;
cl_command_queue queue;
cl_int ret_code;
cl_kernel md4_crypt_kernel;
cl_mem pinned_saved_keys, pinned_partial_hashes, buffer_out, buffer_keys;
static cl_uint *partial_hashes;
static char *saved_plain;
//static int max_key_length = 0;
static char get_key_saved[PLAINTEXT_LENGTH+1];
#define MIN_KEYS_PER_CRYPT      MD4_NUM_KEYS
#define MAX_KEYS_PER_CRYPT      MD4_NUM_KEYS
static size_t global_work_size = MD4_NUM_KEYS;
static size_t local_work_size;

#else

#define PLAINTEXT_LENGTH        125
#define MIN_KEYS_PER_CRYPT      1
#define MAX_KEYS_PER_CRYPT      1
static int saved_key_length;
static char saved_key[PLAINTEXT_LENGTH + 1];
static MD4_CTX ctx;
static ARCH_WORD_32 crypt_out[4];

#endif

static struct fmt_tests tests[] = {
{"bde52cb31de33e46245e05fbdbd6fb24", "a"},
{"8a9d093f14f8701df17732b2bb182c74", "password"},
{"$MD4$8a9d093f14f8701df17732b2bb182c74", "password"},
{"f2756eedf49246eefaf489d2a6263f13", "slashdot"},
{NULL}
};

#ifdef CL_VERSION_1_0

static void if_error_log(cl_int ret_code, const char *message)
{
    if(ret_code != CL_SUCCESS) {
        printf("\nOpenCL: %s\n", message);
        exit(-1);
    }
}

// TODO: free resources at exit
static void fmt_MD4_init(void)
{
    // load kernel source
    char *source=(char*)mem_alloc(1024*16);
    FILE *fp = fopen(path_expand("$JOHN/md4_opencl_kernel.cl"),"r");
    if(!fp)
        if_error_log(!CL_SUCCESS, "Source kernel not found!");
    size_t source_size = fread(source, sizeof(char), 1024*16, fp);
    source[source_size] = 0;
    fclose(fp);

    // get a platform and its information
    size_t max_group_size;
    char log[1024*64];
    ret_code = clGetPlatformIDs(1, &platform, NULL);
    if_error_log(ret_code, "No OpenCL platform exist");
    ret_code = clGetPlatformInfo(platform, CL_PLATFORM_NAME, sizeof(log), log, NULL);
    if_error_log(ret_code, "Error querying PLATFORM_NAME");
    printf("\nOpenCL Platform: <<<%s>>>", log);

    // find an OpenCL device
    //ret_code = clGetDeviceIDs(platform, CL_DEVICE_TYPE_CPU, 1, &devices, NULL); // doesn't work on NVIDIA cards
    ret_code = clGetDeviceIDs(platform, CL_DEVICE_TYPE_GPU, 1, &devices, NULL);
    if_error_log(ret_code, "No OpenCL device of that type exist");
    ret_code = clGetDeviceInfo(devices, CL_DEVICE_NAME, sizeof(log), log, NULL);
    if_error_log(ret_code, "Error querying DEVICE_NAME");
    printf(" and device: <<<%s>>>\n",log);
    ret_code = clGetDeviceInfo(devices, CL_DEVICE_MAX_WORK_GROUP_SIZE, sizeof(max_group_size), &max_group_size, NULL);
    if_error_log(ret_code, "Error querying MAX_WORK_GROUP_SIZE");

    // create a context and command queue on the device.
    context = clCreateContext(NULL, 1, &devices, NULL, NULL, &ret_code);
    if_error_log(ret_code, "Error creating context");
    queue = clCreateCommandQueue(context, devices, 0, &ret_code);
    if_error_log(ret_code, "Error creating command queue");

    // submit the kernel source for compilation
    program = clCreateProgramWithSource(context, 1, (const char **)&source, NULL, &ret_code);
    if_error_log(ret_code,"Error creating program");
    ret_code = clBuildProgram(program, 1, &devices, NULL, NULL, NULL);
    if(ret_code != CL_SUCCESS) {
        printf("failed in clBuildProgram with %d\n", ret_code);
        clGetProgramBuildInfo(program, devices, CL_PROGRAM_BUILD_LOG, sizeof(log), (void*)log, NULL);
        printf("compilation log: %s\n", log);
        exit(-1);
    }

    // create kernel to execute
    md4_crypt_kernel = clCreateKernel(program, "md4", &ret_code);
    if_error_log(ret_code, "Error creating kernel");

    // create Page-Locked (Pinned) memory for higher bandwidth between host and device (Nvidia Best Practices)
    pinned_saved_keys = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR,
            (PLAINTEXT_LENGTH+1)*MD4_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating page-locked memory");
    pinned_partial_hashes = clCreateBuffer(context, CL_MEM_READ_WRITE | CL_MEM_ALLOC_HOST_PTR, 4*MD4_NUM_KEYS, NULL, &ret_code);

    if_error_log (ret_code, "Error creating page-locked memory");
    saved_plain = (char*)clEnqueueMapBuffer(queue, pinned_saved_keys, CL_TRUE, CL_MAP_WRITE | CL_MAP_READ, 0,
            (PLAINTEXT_LENGTH+1)*MD4_NUM_KEYS, 0, NULL, NULL, &ret_code);
    if_error_log (ret_code, "Error mapping page-locked memory");
    partial_hashes = (cl_uint*)clEnqueueMapBuffer(queue, pinned_partial_hashes, CL_TRUE, CL_MAP_READ, 0, 4*MD4_NUM_KEYS, 0, NULL, NULL, &ret_code);
    if_error_log (ret_code, "Error mapping page-locked memory");

    // create and set arguments
    buffer_keys = clCreateBuffer(context, CL_MEM_READ_ONLY, (PLAINTEXT_LENGTH+1)*MD4_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code, "Error creating buffer argument");
    buffer_out  = clCreateBuffer(context, CL_MEM_WRITE_ONLY, (BINARY_SIZE)*MD4_NUM_KEYS, NULL, &ret_code);
    if_error_log (ret_code,"Error creating buffer argument");

    ret_code = clSetKernelArg(md4_crypt_kernel, 0, sizeof(buffer_keys), (void*) &buffer_keys);
    if_error_log (ret_code, "Error setting argument 1");
    ret_code = clSetKernelArg(md4_crypt_kernel, 1, sizeof(buffer_out ), (void*) &buffer_out);
    if_error_log (ret_code, "Error setting argument 2");

    local_work_size = 64; // TODO: detect dynamically
}

#endif

static int valid(char *ciphertext)
{
    char *p, *q;

    p = ciphertext;
    if (!strncmp(p, "$MD4$", 5))
        p += 5;

    q = p;
    while (atoi16[ARCH_INDEX(*q)] != 0x7F)
        q++;
    return !*q && q - p == CIPHERTEXT_LENGTH;
}

static char *split(char *ciphertext, int index)
{
    static char out[5 + CIPHERTEXT_LENGTH + 1];

    if (!strncmp(ciphertext, "$MD4$", 5))
        return ciphertext;

    memcpy(out, "$MD4$", 5);
    memcpy(out + 5, ciphertext, CIPHERTEXT_LENGTH + 1);
    return out;
}

static void *get_binary(char *ciphertext)
{
    static unsigned char out[BINARY_SIZE];
    char *p;
    int i;

    p = ciphertext + 5;
    for (i = 0; i < sizeof(out); i++) {
        out[i] =
            (atoi16[ARCH_INDEX(*p)] << 4) |
            atoi16[ARCH_INDEX(p[1])];
        p += 2;
    }

    return out;
}

static int binary_hash_0(void *binary)
{
    return *(ARCH_WORD_32 *)binary & 0xF;
}

static int binary_hash_1(void *binary)
{
    return *(ARCH_WORD_32 *)binary & 0xFF;
}

static int binary_hash_2(void *binary)
{
    return *(ARCH_WORD_32 *)binary & 0xFFF;
}

static int binary_hash_3(void *binary)
{
    return *(ARCH_WORD_32 *)binary & 0xFFFF;
}

static int binary_hash_4(void *binary)
{
    return *(ARCH_WORD_32 *)binary & 0xFFFFF;
}

static int get_hash_0(int index)
{
#ifdef CL_VERSION_1_0
#ifdef DEBUG
    printf("* in get_hash0, index : %d, hash : ", index);
    int i;
    for(i = 0; i < 4; i++)
        printf("%02x ", partial_hashes[i*MD4_NUM_KEYS + index]);
    printf("\n");
#endif
    return partial_hashes[index] & 0x0F;
#else
    return crypt_out[0] & 0xF;
#endif
}

static int get_hash_1(int index)
{
#ifdef CL_VERSION_1_0
    return partial_hashes[index] & 0xFF;
#else
    return crypt_out[0] & 0xFF;
#endif
}

static int get_hash_2(int index)
{
#ifdef CL_VERSION_1_0
    return partial_hashes[index] & 0xFFF;
#else
    return crypt_out[0] & 0xFFF;
#endif
}

static int get_hash_3(int index)
{
#ifdef CL_VERSION_1_0
    return partial_hashes[index] & 0xFFFF;
#else
    return crypt_out[0] & 0xFFFF;
#endif
}

static int get_hash_4(int index)
{
#ifdef CL_VERSION_1_0
    return partial_hashes[index] & 0xFFFFF;
#else
    return crypt_out[0] & 0xFFFFF;
#endif
}

static void set_salt(void *salt)
{
}

#ifdef CL_VERSION_1_0
static void set_key(char *key, int index)
{
    int length = -1;
    int base = index * (PLAINTEXT_LENGTH+1);
    do
    {
        length++;
        saved_plain[base + length] = key[length];
    }
    while(key[length]);
    memset(&saved_plain[base + length+1], 0, 7); // ugly hack which "should" work!
}

static char *get_key(int index)
{
    int length = -1;
    int base = index * (PLAINTEXT_LENGTH+1);
    do
    {
        length++;
        get_key_saved[length] = saved_plain[base + length];
    }
    while(get_key_saved[length]);
    get_key_saved[length]= 0;
    return get_key_saved;
}

static void crypt_all(int count)
{
#ifdef DEBUGVERBOSE
    int i, j;
    unsigned char *p = (unsigned char*)saved_plain;
    count--;
    for(i=0; i<count+1; i++) {
        printf("\npassword : ");
        for(j=0; j < 64; j++) {
            printf("%02x ", p[i*64 + j]);
        }
    }
    printf("\n");
#endif
    cl_int code;
    // copy keys to the device
    code = clEnqueueWriteBuffer(queue, buffer_keys, CL_TRUE, 0, (PLAINTEXT_LENGTH+1) * MD4_NUM_KEYS, saved_plain, 0, NULL, NULL);
    if(code != CL_SUCCESS) {
        printf("failed in clEnqueueWriteBuffer with code %d\n", code);
        exit(-1);
    }
    // execute md4 kernel
    code = clEnqueueNDRangeKernel(queue, md4_crypt_kernel, 1, NULL, &global_work_size, &local_work_size, 0, NULL, NULL);
    if(code != CL_SUCCESS) {
        printf("failed in clEnqueueNDRangeKernel with code %d\n", code);
        exit(-1);
    }
    clFinish(queue);
    // read back partial hashes
    clEnqueueReadBuffer(queue, buffer_out, CL_TRUE, 0, 4*MD4_NUM_KEYS, partial_hashes, 0, NULL, NULL);

#ifdef DEBUGVERBOSE
    p = (unsigned char *)partial_hashes;
    for(i=0; i<2; i++) {
        printf("\n\npartial_hashes : ");
        for(j=0; j < 16; j++)
            printf("%02x ", p[i*16 + j]);
    }
    printf("\n");;
#endif
}

/* 1/4 hash data hack forces this "usually unlikely" function to be slightly more complex */
static int cmp_one(void * binary, int index)
{
    unsigned int *t=(unsigned int *)binary;
    unsigned int a;
    unsigned int c;
    unsigned int d;

    // b
    clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(1*MD4_NUM_KEYS+index), sizeof(a), (void*)&a, 0, NULL, NULL);
    if (t[1]!=a)
        return 0;
    // c
    clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(2*MD4_NUM_KEYS+index), sizeof(c), (void*)&c, 0, NULL, NULL);
    if (t[2]!=c)
        return 0;
    // d
    clEnqueueReadBuffer(queue, buffer_out, CL_TRUE,  sizeof(cl_uint)*(3*MD4_NUM_KEYS+index), sizeof(d), (void*)&d, 0, NULL, NULL);
    return t[3]==d;
}

static int cmp_all(void *binary, int count)
{
#ifdef DEBUG
    printf("* in cmp_all with count %d\n", count);
#endif
    unsigned int i=0;
    unsigned int b=((unsigned int *)binary)[0];
    for(; i<count; i++)
         if(b==partial_hashes[i])
             return 1;
    return 0;
}

#else

static void set_key(char *key, int index)
{
    saved_key_length = strlen(key);
    if (saved_key_length > PLAINTEXT_LENGTH)
        saved_key_length = PLAINTEXT_LENGTH;
    memcpy(saved_key, key, saved_key_length);
}

static char *get_key(int index)
{
    saved_key[saved_key_length] = 0;
    return saved_key;
}

static void crypt_all(int count)
{
    MD4_Init(&ctx);
    MD4_Update(&ctx, saved_key, saved_key_length);
    MD4_Final((unsigned char *)crypt_out, &ctx);
}

static int cmp_all(void *binary, int count)
{
    return !memcmp(binary, crypt_out, BINARY_SIZE);
}
#endif

static int cmp_exact(char *source, int index)
{
    return 1;
}

struct fmt_main fmt_rawMD4 = {
    {
        FORMAT_LABEL,
        FORMAT_NAME,
        ALGORITHM_NAME,
        BENCHMARK_COMMENT,
        BENCHMARK_LENGTH,
        PLAINTEXT_LENGTH,
        BINARY_SIZE,
        SALT_SIZE,
        MIN_KEYS_PER_CRYPT,
        MAX_KEYS_PER_CRYPT,
        FMT_CASE | FMT_8_BIT,
        tests
    }, {
#ifdef CL_VERSION_1_0
        fmt_MD4_init,
#else
        fmt_default_init,
#endif
        valid,
        split,
        get_binary,
        fmt_default_salt,
        {
            binary_hash_0,
            binary_hash_1,
            binary_hash_2,
            binary_hash_3,
            binary_hash_4
        },
        fmt_default_salt_hash,
        set_salt,
        set_key,
        get_key,
        fmt_default_clear_keys,
        crypt_all,
        {
            get_hash_0,
            get_hash_1,
            get_hash_2,
            get_hash_3,
            get_hash_4
        },
        cmp_all,
#ifdef CL_VERSION_1_0
        cmp_one,
#else
        cmp_all,
#endif
        cmp_exact
    }
};
